#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "xmms2.h"
#include "util.h"

extern int hsc_mode;
extern GtkWidget *mainwindow;
extern GtkWidget *infowindow;
extern GtkWidget *volumewindow;
extern GtkWidget *morewindow;
extern plsnfo my_playlist;

void on_cbtn_connect_toggled           (GtkToggleButton *togglebutton,
                                        gpointer         user_data) {
  if(gtk_toggle_button_get_active (togglebutton)) {
    init_xmms_session();
  } else {
    close_xmms_session();
  }
  update_gui();
}


void on_xtr_prev_activate              (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_PREV);
}


void on_xtr_restart_activate           (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_CURR);
}


void on_xtr_back_activate              (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_FRWD);
}


void on_xtr_play_activate              (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_PLAY);
}


void on_xtr_pause_activate             (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_PAUSE);
}


void on_xtr_stop_activate              (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_STOP);
}


void on_xtr_forw_activate              (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_FFWD);
}


void on_xtr_next_activate              (GtkButton       *button,
                                        gpointer         user_data) {
  xmms2_cmd(XCMD_NEXT);
}

void on_hsc_seek_value_changed         (GtkRange        *range,
                                        gpointer         user_data) {
  gdouble val = gtk_range_get_value (range);
  if (hsc_mode == 0 ) { return;}
  xmms_reposition(val);
  //printf("reposition : val %.3f\n",val);
  if (hsc_mode == 2 ) {hsc_mode=0;}
}


void on_hsc_seek_move_slider           (GtkRange        *range,
                                        GtkScrollType    scroll,
                                        gpointer         user_data) {
  hsc_mode=2;
}


gboolean on_hsc_seek_change_value      (GtkRange        *range,
                                        GtkScrollType    scroll,
                                        gdouble          value,
                                        gpointer         user_data) {
  return FALSE;
}


gboolean on_hsc_seek_button_press_event(GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data) {
  hsc_mode=1;
  return FALSE;
}


gboolean on_hsc_seek_button_release_event (GtkWidget       *widget,
                                           GdkEventButton  *event,
                                           gpointer         user_data) {
  hsc_mode=0;
  return FALSE;
}

void on_row_activated                  (GtkTreeView     *treeview,
                                        GtkTreePath     *path,
                                        GtkTreeViewColumn *column,
                                        gpointer         user_data) {
  xmms_pljump(position_in_tree(gtk_tree_view_get_model(treeview),path), true);
}


void on_btn_del_activate               (GtkButton       *button,
                                        gpointer         user_data) {
  // remove current selection
  GtkTreeSelection *ts = NULL;
  ts = gtk_tree_view_get_selection(GTK_TREE_VIEW(lookup_widget(mainwindow,"treeview1")));
  gtk_tree_selection_selected_foreach(ts, del_sel_foreach, NULL);
}


void on_btn_about_activate             (GtkButton       *button,
                                        gpointer         user_data) {
  status_printf("gx2ctl %s - robin@gareus.org",VERSION);
  GtkWidget *about = create_aboutdialog1();
  gtk_dialog_run (GTK_DIALOG (about));
  gtk_widget_destroy (about);
}


void on_gx2ctl_destroy                 (GtkObject       *object,
                                        gpointer         user_data) {
  printf("quit!\n");
  close_xmms_session();
  gtk_main_quit();
}


void
on_btn_info_activate                   (GtkButton       *button,
                                        gpointer         user_data) {
  gtk_widget_hide (infowindow);
  gtk_widget_show (infowindow);
}


void
on_btn_vol_activate                    (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide (volumewindow);
  gtk_widget_show (volumewindow);
}


char *remember_filename = NULL;

void
on_btn_add_activate                    (GtkButton       *button,
                                        gpointer         user_data) {
  GtkWidget *dialog = create_filechooserdialog1();
  if (remember_filename)
    gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog),remember_filename);

  // TODO don't run dialog in fg.  -> callback on add.
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK) {
    char *filename;
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
    if(remember_filename) free(remember_filename);
    remember_filename=strdup(filename);
    fprintf(stdout,"open file: %s\n",filename);
    add_to_playlist(filename,-1);
    g_free (filename);
//} else {
//  fprintf(stdout,"I ain't open no files just now.\n");
  }
  gtk_widget_destroy (dialog);
}

void
on_btn_more_activate                   (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide (morewindow);
  gtk_widget_show (morewindow);
}


void
on_be_clear_activate                   (GtkButton       *button,
                                        gpointer         user_data)
{
  xmms2_cmd(XCMD_CLEAR);
}


void
on_be_append_activate                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *w= lookup_widget(morewindow,"e_url");
  add_to_playlist ((char*)gtk_entry_get_text(GTK_ENTRY(w)),-1);
  //gtk_entry_set_text(GTK_ENTRY(w),"");
}


void
on_be_insert_activate                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *w= lookup_widget(morewindow,"e_url");
  add_to_playlist ((char*)gtk_entry_get_text(GTK_ENTRY(w)), (gint) my_playlist.playlist_pos);
}


void
on_e_wipe_activate                     (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *w= lookup_widget(morewindow,"e_url");
  gtk_entry_set_text(GTK_ENTRY(w),"");
}


gboolean
on_tvp_button_press_event              (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  gboolean result = FALSE;
    if (event->type == GDK_BUTTON_PRESS  &&  event->button == 3) {
    setup_colmenu (widget, event, user_data);
    result = TRUE;
  }
  return result;
}


void
on_be_refresh_activate                 (GtkButton       *button,
                                        gpointer         user_data)
{
  refresh_playlist();
}

/* vi:set ts=8 sts=2 sw=2: */
