/* 
   Copyright (C) 2007 Robin Gareus <robin@gareus.org>

   URI string handlers from sdecode, gxmms2 by Johannes Heimansberg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/
//<-TERMINAL 124---------------------------------------------------------------------------------------------------------->/
#define _GNU_SOURCE

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "support.h"
#include "xmms2.h"
#include "util.h"

extern GtkWidget *mainwindow;
extern GtkWidget *infowindow;
extern GtkWidget *volumewindow;
extern plsnfo my_playlist;

GtkListStore *lsp; ///< playlist list store

/*
 * status bar
 */

void status_msg (char *msg) {
	gchar buf[PATH_MAX];
	guint context = 1;
	GtkWidget *sb;
	sb=lookup_widget(mainwindow,"statusbar");

	snprintf(buf, PATH_MAX, "%s", msg);
	gtk_statusbar_pop(GTK_STATUSBAR(sb), context);
	gtk_statusbar_push(GTK_STATUSBAR(sb), context,  buf);
}

void status_printf (const char *format, ...) {
	va_list arglist;
	char text[PATH_MAX];

	va_start(arglist, format);
	vsnprintf(text, PATH_MAX, format, arglist);
	va_end(arglist);
	text[PATH_MAX -1] =0;
	
	status_msg(text);
}

/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * string helpers
 */

gchar *get_valid_utf8_str_alloc(const gchar *str_to_validate) {
  gchar *result = g_malloc(strlen(str_to_validate) + 1);

  if (g_utf8_validate(str_to_validate, -1, NULL)) {
    g_strlcpy(result, str_to_validate, strlen(str_to_validate) + 1);
  } else {
    int i;
    for (i = 0; i < strlen(str_to_validate) + 1; i++) {
      if (str_to_validate[i] >= 0) /* negative chars are non-ASCII (>127) chars */
	result[i] = str_to_validate[i];
      else
	result[i] = '?';
    }
  }
  return result;
}

gchar hex2char(gchar first_digit, gchar second_digit) {
	gchar result = 0;
	
	if (first_digit >= '0' && first_digit <= '9')
		result = (first_digit - '0') * 16;
	else if (first_digit >= 'a' && first_digit <= 'f')
		result = (first_digit - 'a' + 10) * 16;
	if (second_digit >= '0' && second_digit <= '9')
		result += second_digit - '0';
	else if (second_digit >= 'a' && second_digit <= 'f')
		result += second_digit - 'a' + 10;
	return result;
}

gchar *decode_string(gchar *string) {
	int i, j;
	gchar *result = g_malloc((strlen(string) + 1) * sizeof(gchar));

	/* decode %XX escape sequences to their corresponding chars: */
	for (i = 0, j = 0; i < strlen(string); i++) {
		if (string[i] == '%') {
			result[j] = hex2char(string[i+1], string[i+2]);
			i+=2;
			j++; 
		} else {
			result[j] = string[i];
			j++;
		}
	}
	result[j] = '\0';
	return result;
}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * DND
 */

#define TGNR(a) (guint)(sizeof(a)/sizeof(GtkTargetEntry))
/* Drag and drop types for search & song tables */
enum { DND_TEXT_URI_LIST=1000, DND_TEXT_PLAIN, DND_PLINTERNAL };

static GtkTargetEntry drag_types [] = { 
	{ "text/uri-list", 0, DND_TEXT_URI_LIST }, 
	{ "text/plain", 0, DND_TEXT_PLAIN }, 
	{ "gxlist", 0, DND_PLINTERNAL }, 
	{ "STRING", 0, DND_TEXT_PLAIN }
};

static GtkTargetEntry drop_trash [] = { 
	{ "gxlist", 0, DND_PLINTERNAL }
};

static GtkTargetEntry drop_add [] = { 
	{ "text/uri-list", 0, DND_TEXT_URI_LIST }, 
	//{ "text/x-moz-url", 0, DND_TEXT_URI_LIST },  //< UTF16 ??
	{ "application/x-rootwindow-drop", 0, DND_TEXT_PLAIN }
};

static GtkTargetEntry drop_types [] = { 
	{ "gxlist", 0, DND_PLINTERNAL },
	{ "text/uri-list", 0, DND_TEXT_URI_LIST }, 
	//{ "text/x-moz-url", 0, DND_TEXT_URI_LIST },  //< UTF16 ??
	{ "application/x-rootwindow-drop", 0, DND_TEXT_PLAIN }
};

static void
on_tm_dnd_get_pli_foreach(GtkTreeModel *tm, GtkTreePath *tp,
			  GtkTreeIter *iter, gpointer data) {
  gchar *url;
  GString *repl = (GString *)data;

  GtkTreePath *gtp=gtk_tree_model_get_path(tm,iter);
  g_string_append_printf (repl, "%u\n", position_in_tree(tm,gtp)); 
  gtk_tree_path_free(gtp);
}


static void
on_tm_dnd_get_uri_foreach(GtkTreeModel *tm, GtkTreePath *tp,
			  GtkTreeIter *iter, gpointer data) {
  gchar *url;
  GString *filelist = (GString *)data;

  gtk_tree_model_get(tm, iter, 0, &url, -1);
  gchar *uri = g_filename_to_uri (url, NULL, NULL);
  g_string_append_printf (filelist, "%s\n", url); // XXX uri
  g_free (uri);
}


static void tm_drag_data_get (GtkWidget       *widget,
			      GdkDragContext  *context,
			      GtkSelectionData *data,
			      guint            info,
			      guint            time,
			      gpointer         user_data) {
    GtkTreeSelection *ts = NULL;
    GString *reply = g_string_sized_new (2047);

    if((data) && (ts = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget))))
    {
	switch (info)
	{
	case DND_TEXT_URI_LIST:
	case DND_TEXT_PLAIN:
	    gtk_tree_selection_selected_foreach(ts, on_tm_dnd_get_uri_foreach, reply);
	    break;
	case DND_PLINTERNAL:
	    gtk_tree_selection_selected_foreach(ts, on_tm_dnd_get_pli_foreach, reply);
	    break;
	default:
	    g_warning ("Programming error: ta_drag_data_get received unknown info type (%d)\n", info);
	    break;
	}
    }
    gtk_selection_data_set(data, data->target, 8, (guchar*) reply->str, reply->len);
    g_string_free (reply, TRUE);
}

void cb_drag_add_received(GtkWidget          *widget,
		          GdkDragContext     *context,
		          gint                x,
		          gint                y,
		          GtkSelectionData   *data,
		          guint               info,
		          guint               time) {

  if ((data->length >= 0) && (data->format == 8)) {
    gint n = 0, m = 0;
    gchar buffer[384], *uri_conv;
    /* seperate each url in the list */
    while (n < strlen((gchar *)data->data)) {
      if (((guchar *)data->data)[n] == '\n') {
	buffer[m] = '\0';
	m = 0;
	/* decode %XX escape sequences to their corresponding chars: */
	uri_conv = decode_string(buffer);
	add_to_playlist (uri_conv, -1);
	g_free(uri_conv);
      } else if (((guchar *)data->data)[n] > 31) {
	buffer[m] = ((guchar *)data->data)[n];
	if (m < 382) m++;
      }
      n++;
    }
    gtk_drag_finish(context, TRUE, FALSE, time);
  } else {
    gtk_drag_finish(context, FALSE, FALSE, time);
  }
}


void cb_drag_data_received(GtkWidget          *widget,
		           GdkDragContext     *context,
		           gint                x,
		           gint                y,
		           GtkSelectionData   *data,
		           guint               info,
		           guint               time) {
  GtkTreeViewDropPosition pos = 0;
  guint dest = 0;
  GtkTreePath *path;

  bool destexists = gtk_tree_view_get_dest_row_at_pos( GTK_TREE_VIEW( widget ), x, y, &path, &pos );
  if (destexists)  {
    gchar *tmp = gtk_tree_path_to_string(path);
    //printf("DROP AT %s  pm:%i \n",tmp,pos);
    dest = atoi(tmp); 
    if (pos&2) dest++;
    g_free(tmp);
  } else {
    //printf("DROP at end.\n");
    if (my_playlist.playlist_length>0)
      dest=my_playlist.playlist_length-1;
    else dest=0;
  }

  if ((data->length >= 0) && (data->format == 8)) {

    // TODO: if !destexits -> drop at END!
    if (info == DND_PLINTERNAL) {
      gchar *s= (gchar*)data->data,*p; 
      while (s && (p=strchr(s,'\n'))) {
	printf("DND-MOVE # %u -> %u\n",atoi(s),dest);
	xmms_plmv(atoi(s),dest);
	s=++p;
      }
      return;
    }

    gint n = 0, m = 0;
    gchar buffer[384], *uri_conv;
    /* seperate each url in the list */
    while (n < strlen((gchar *)data->data)) {
      if (((guchar *)data->data)[n] == '\n') {
	buffer[m] = '\0';
	m = 0;
	/* decode %XX escape sequences to their corresponding chars: */
	uri_conv = decode_string(buffer);
	if (destexists) add_to_playlist (uri_conv, dest);
	else add_to_playlist (uri_conv, -1);
	g_free(uri_conv);
      } else if (((guchar *)data->data)[n] > 31) {
	buffer[m] = ((guchar *)data->data)[n];
	if (m < 382) m++;
      }
      n++;
    }
    gtk_drag_finish(context, TRUE, FALSE, time);
  } else {
    gtk_drag_finish(context, FALSE, FALSE, time);
  }
}

void cb_drag_trash_received(GtkWidget          *widget,
		            GdkDragContext     *context,
		            gint                x,
		            gint                y,
		            GtkSelectionData   *data,
		            guint               info,
		            guint               time) {
  GtkTreeViewDropPosition pos;
  GtkTreePath *path;

	if ((data->length >= 0) && (data->format == 8)) {
		gchar *s= (gchar*)data->data,*p; 
		while (s && (p=strchr(s,'\n'))) {
		  printf("DND-RM #%u\n",atoi(s));
		  xmms_plrm((guint)(atoi(s)));
		  s=++p;
		}
		gtk_drag_finish(context, TRUE, FALSE, time);
	} else {
		gtk_drag_finish(context, FALSE, FALSE, time);
	}
}


static void pl_cell_data_function(GtkTreeViewColumn *col,
				  GtkCellRenderer   *renderer,
				  GtkTreeModel      *model,
				  GtkTreeIter       *iter,
				  gpointer           userdata) {
  GtkTreePath *gtp=gtk_tree_model_get_path(model,iter);
  guint pos = position_in_tree(model,gtp);
  gtk_tree_path_free(gtp);
  if (pos == my_playlist.playlist_pos) 
    g_object_set(renderer, "weight", PANGO_WEIGHT_BOLD, "weight-set", TRUE, NULL);
  else
    g_object_set(renderer, "weight-set", FALSE, NULL);
}

void ls_update(GtkTreeIter *iter, trackinfo *ti){
  gchar *tplt = trackinfo_get_playtime_str_alloc(ti);
  gchar tttl[256];
  if (trackinfo_is_stream(ti) && trackinfo_has_title(ti)) {
    g_snprintf(tttl, sizeof(tttl), "[stream] %s", trackinfo_get_title_str(ti));
  } else if (trackinfo_has_title(ti)) {
    g_snprintf(tttl, sizeof(tttl), "%s", trackinfo_get_title_str(ti));
  } else {
    gchar *filename;
    filename = g_path_get_basename (trackinfo_get_url_str(ti));
    if (filename) {
      g_snprintf (tttl, sizeof (tttl), "%s", filename);
      g_free (filename);
    } else {
      g_snprintf(tttl, sizeof(tttl), " ~ UNKNOWN ~ "); /// XXX full URL?
    }
  }

  gtk_list_store_set(lsp, iter, 0, trackinfo_get_url_str(ti),
				1, tttl,
				2, trackinfo_get_artist_str(ti),
				3, trackinfo_get_album_str(ti),
				4, trackinfo_get_title_str(ti),
				5, tplt,
			       -1);
  g_free(tplt);
  //g_free(tttl);
}

/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * Playlist Treeview
 */


#define LAST_COLUMN (6) 

void setup_treeview (void) {
  GtkTreeView *tvp = GTK_TREE_VIEW(lookup_widget(mainwindow, "treeview1")); 
  GtkTreeViewColumn *col;
  GtkCellRenderer *renderer;

  lsp = gtk_list_store_new(LAST_COLUMN, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

  gtk_tree_view_set_model (tvp, GTK_TREE_MODEL (lsp));
  //gtk_tree_view_set_headers_clickable(tvp, TRUE);

#if 1
  gtk_tree_selection_set_mode (gtk_tree_view_get_selection (tvp), GTK_SELECTION_BROWSE);
#else
  gtk_tree_selection_set_mode (gtk_tree_view_get_selection (tvp), GTK_SELECTION_MULTIPLE);
#endif 

  renderer = gtk_cell_renderer_text_new();
  col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col, "URL");
  gtk_tree_view_append_column(tvp, col);
  gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_AUTOSIZE);
  gtk_tree_view_column_pack_start(col, renderer, TRUE);
  gtk_tree_view_column_add_attribute(col, renderer, "text", 0);
  gtk_tree_view_column_set_resizable(col, TRUE);
  gtk_tree_view_column_set_cell_data_func(col, renderer, pl_cell_data_function, lsp, NULL);
  gtk_tree_view_column_set_visible(col,FALSE);
  gtk_tree_view_column_set_reorderable(col,TRUE);

  col = gtk_tree_view_column_new_with_attributes("File", renderer, "text", 1, NULL);
  gtk_tree_view_append_column(tvp, col);
  gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED); // XXX
  gtk_tree_view_column_set_resizable(col, TRUE);
  gtk_tree_view_column_set_expand(col, TRUE);
  gtk_tree_view_column_set_cell_data_func(col, renderer, pl_cell_data_function, lsp, NULL);
  gtk_tree_view_column_set_reorderable(col,TRUE);

  col = gtk_tree_view_column_new_with_attributes("Artist", renderer, "text", 2, NULL);
  gtk_tree_view_append_column(tvp, col);
  gtk_tree_view_column_set_resizable(col, TRUE);
  gtk_tree_view_column_set_expand(col, TRUE);
  gtk_tree_view_column_set_cell_data_func(col, renderer, pl_cell_data_function, lsp, NULL);
  gtk_tree_view_column_set_visible(col,FALSE);
  gtk_tree_view_column_set_reorderable(col,TRUE);

  col = gtk_tree_view_column_new_with_attributes("Album", renderer, "text", 3, NULL);
  gtk_tree_view_append_column(tvp, col);
  gtk_tree_view_column_set_resizable(col, TRUE);
  gtk_tree_view_column_set_expand(col, TRUE);
  gtk_tree_view_column_set_cell_data_func(col, renderer, pl_cell_data_function, lsp, NULL);
  gtk_tree_view_column_set_visible(col,FALSE);
  gtk_tree_view_column_set_reorderable(col,TRUE);

  col = gtk_tree_view_column_new_with_attributes("Title", renderer, "text", 4, NULL);
  gtk_tree_view_append_column(tvp, col);
  gtk_tree_view_column_set_resizable(col, TRUE);
  gtk_tree_view_column_set_expand(col, TRUE);
  gtk_tree_view_column_set_cell_data_func(col, renderer, pl_cell_data_function, lsp, NULL);
  gtk_tree_view_column_set_visible(col,FALSE);
  gtk_tree_view_column_set_reorderable(col,TRUE);

  col = gtk_tree_view_column_new_with_attributes("Time", renderer, "text", 5, NULL);
  gtk_tree_view_append_column(tvp, col);
  gtk_tree_view_column_set_sizing(col,GTK_TREE_VIEW_COLUMN_FIXED); // XXX
  gtk_tree_view_column_set_resizable(col, TRUE);
  gtk_tree_view_column_set_expand(col, TRUE);
  gtk_tree_view_column_set_cell_data_func(col, renderer, pl_cell_data_function, lsp, NULL);
  gtk_tree_view_column_set_reorderable(col,TRUE);


  // set up DnD 
  gtk_drag_source_set(GTK_WIDGET(tvp),GDK_BUTTON1_MASK, drag_types, TGNR(drag_types), GDK_ACTION_MOVE);
  g_signal_connect ((gpointer) tvp, "drag_data_get", G_CALLBACK(tm_drag_data_get),NULL);

  gtk_drag_dest_set(GTK_WIDGET(tvp), GTK_DEST_DEFAULT_ALL, drop_types, TGNR(drop_types), GDK_ACTION_COPY| GDK_ACTION_MOVE);
  g_signal_connect ((gpointer) tvp, "drag_data_received", G_CALLBACK (cb_drag_data_received), NULL);

  //setup trash drop
  GtkWidget *trash = lookup_widget(mainwindow, "btn_del"); 
  gtk_drag_dest_set(trash, GTK_DEST_DEFAULT_ALL, drop_trash, TGNR(drop_trash), GDK_ACTION_MOVE);
  g_signal_connect ((gpointer) trash, "drag_data_received", G_CALLBACK (cb_drag_trash_received), NULL);

  //setup add drop
  GtkWidget *add = lookup_widget(mainwindow, "btn_add"); 
  gtk_drag_dest_set(GTK_WIDGET(add), GTK_DEST_DEFAULT_ALL, drop_add, TGNR(drop_add), GDK_ACTION_COPY| GDK_ACTION_MOVE);
  g_signal_connect ((gpointer) add, "drag_data_received", G_CALLBACK (cb_drag_add_received), NULL);

}

gchar *get_col_title(gint col) {
  GtkTreeView *tvp = GTK_TREE_VIEW(lookup_widget(mainwindow, "treeview1")); 
  GtkTreeViewColumn *c = gtk_tree_view_get_column (tvp, col);
  return ((gchar*) gtk_tree_view_column_get_title(c));
}

gboolean is_col_displayed(gint col) {
  GtkTreeView *tvp = GTK_TREE_VIEW(lookup_widget(mainwindow, "treeview1")); 
  GtkTreeViewColumn *c = gtk_tree_view_get_column (tvp, col);
  return (gtk_tree_view_column_get_visible(c));
}

static void column_callback_fn (GtkWidget *menuitem,
                                gpointer   userdata) {
  gint col = (gint) userdata;
  GtkTreeView *tvp = GTK_TREE_VIEW(lookup_widget(mainwindow, "treeview1")); 
  GtkTreeViewColumn *c = gtk_tree_view_get_column (tvp, col);
  gtk_tree_view_column_set_visible(c, !(gtk_tree_view_column_get_visible(c)));
}

void setup_colmenu (GtkWidget      *treeview,
		    GdkEventButton *event,
		    gpointer        userdata) {
  GtkWidget *menuitem, *menu_columns = gtk_menu_new();

  gint i = 0;
  while (i<LAST_COLUMN) { 
    menuitem = gtk_check_menu_item_new_with_mnemonic(get_col_title(i));
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem), is_col_displayed(i));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu_columns), menuitem);
    g_signal_connect(menuitem, "activate", G_CALLBACK(column_callback_fn), (void*) i);
    i++;
  }

  gtk_widget_show_all(menu_columns);
  gtk_menu_popup(GTK_MENU(menu_columns), NULL, NULL, NULL, NULL, (event != NULL) ? event->button : 0, gdk_event_get_time((GdkEvent*)event));

}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *  update/refresh 
 */

#define SSENS(WTNAME,VAL) widget=lookup_widget(mainwindow, WTNAME); gtk_widget_set_sensitive (widget, VAL);

void update_status (guint status, gint no_stream) {
  GtkWidget *widget;
  SSENS("xtr_stop",status!=0);
  SSENS("xtr_play",status!=1);
  SSENS("xtr_pause",status!=2);
  SSENS("xtr_forw",((status==1) && (no_stream!=0)));
  SSENS("xtr_back",((status==1) && (no_stream!=0)));
  SSENS("hsc_seek",((status==1) && (no_stream!=0)));
}

void update_gui (void) {
  GtkWidget *widget;
  int xc = xmms_is_connected();
  if (!xc) {
    SSENS("xtr_stop",xc);
    SSENS("xtr_play",xc);
    SSENS("xtr_pause",xc);
    SSENS("xtr_forw",xc);
    SSENS("xtr_back",xc);
    SSENS("hsc_seek",xc);
  }
  SSENS("xtr_next",xc);
  SSENS("xtr_prev",xc);
  SSENS("xtr_restart",xc);
  widget=lookup_widget(mainwindow,"cbtn_connect");
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(widget), xc);
}

#define ULBL(LNAME, STR) w = lookup_widget(infowindow, LNAME ); gtk_label_set_text(GTK_LABEL(w),STR);

#define ALBL(LNAME, ASTR) s = ASTR; w = lookup_widget(infowindow, LNAME ); gtk_label_set_text(GTK_LABEL(w),s);  if(s) g_free(s);
void update_info (trackinfo *t){
  GtkWidget *w;
  gchar *s;
  // if (gtk_is_hidden(infowindow)) return;
  ULBL("l_title", trackinfo_get_title_str(t));
  ULBL("l_album", trackinfo_get_album_str(t)); 
  ULBL("l_artist", trackinfo_get_artist_str(t)); 
  //ULBL("l_rating"); 
  ALBL("l_duration", trackinfo_get_playtime_str_alloc(t)); 
  ALBL("l_samplerate", trackinfo_get_samplerate_str_alloc(t));
  ALBL("l_bitrate", trackinfo_get_bitrate_str_alloc(t));
  ULBL("l_genre", trackinfo_get_genre_str(t)); 
  ULBL("l_date", trackinfo_get_date_str(t)); 
  ULBL("l_channel", trackinfo_get_channel_str(t));
  ULBL("l_url", trackinfo_get_url_str(t));
  ULBL("l_comment", trackinfo_get_comment_str(t)); 
  //ULBL("l_trackno"); 
  //ULBL("l_smpte"); 
  //ULBL("l_remaining"); 
  s=calloc(32,sizeof(gchar));
  g_snprintf(s,sizeof(s),"%u",trackinfo_get_file_size(t));
  ULBL("l_size", s);  g_free(s);
}

void update_infotime (guint tm, guint len){
  unsigned int t = tm/100;
  GtkWidget *w;
  gchar *s;
  s=calloc(32,sizeof(gchar));
  g_snprintf(s,32,"%02u:%02u:%02u.%01u",(t/36000), (t/600)%60, (t/10)%60, t%10);
  ULBL("l_smpte", s);
  if (len>0) { t=(len-tm)/100; } else { t=0; }
  g_snprintf(s,32,"- %02u:%02u:%02u.%01u",(t/36000), (t/600)%60, (t/10)%60, t%10);
  ULBL("l_remaining", s);  g_free(s);
}

void expose_playlist(void) { 
  GtkWidget *tvp = lookup_widget(mainwindow, "treeview1"); 
  gtk_widget_queue_draw(tvp);
}

/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * cover-art & pixbuf's
 */

/** wipe displayed cover-art, clear pixbuf */
void wipe_coverart(void) {
  //printf("DEBUG: clear coverart");
  GtkImage *album_image;
  album_image=GTK_IMAGE(lookup_widget(infowindow,"i_cover"));
  gtk_image_set_from_pixbuf (album_image,NULL);
}

void on_pixbuf_prepared (GdkPixbufLoader *loader, gint width, gint height, gpointer user_data) {
  gdk_pixbuf_loader_set_size(loader,120,120); // TODO: get from prefs.
}

/** set cover image 
 * @param mime: yet unused
 * @param blob: image data
 * @param len: length of image data in bytes
 */
void update_coverart(char *mime, char *blob, size_t len) {
  //printf("DEBUG: coverart %s\n", blob);
  GtkImage *album_image;
  GdkPixbuf *pixbuf;

  album_image=GTK_IMAGE(lookup_widget(infowindow,"i_cover"));
  GdkPixbufLoader *pbl= gdk_pixbuf_loader_new ();
  g_signal_connect ((gpointer) pbl, "size_prepared", G_CALLBACK (on_pixbuf_prepared), NULL );
  gdk_pixbuf_loader_write(pbl, blob, len, NULL);
  pixbuf = gdk_pixbuf_loader_get_pixbuf (pbl);

  if (pixbuf) {
    gtk_image_set_from_pixbuf (album_image,pixbuf);
    g_object_unref(pixbuf);
  }
}

/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * volume stuff
 */

#include "support.h"
static guint initialized_ch = 0;

#define MY_DESTROY_OBJECT(name) \
  g_snprintf(tn, sizeof(tn), name, i); \
  w=lookup_widget(volumewindow,tn); \
  gtk_widget_destroy(w);

#define MY_HOOKUP_OBJECT(component,widget,name) \
  g_snprintf(tn, sizeof(tn), name, i); \
  g_object_set_data_full (G_OBJECT (component), tn, \
    gtk_widget_ref (widget), (GDestroyNotify) gtk_widget_unref); 

void on_vol_value_changed (GtkRange *range, gpointer userdata) {
  GtkWidget *w;
  gdouble val = gtk_range_get_value (range);
  guint ch = (guint) userdata;
  gchar tn[32], *port;

  g_snprintf(tn, sizeof(tn), "voll%i", ch); 
  w=lookup_widget(volumewindow,tn);
  if(gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(w))) {
    // TODO get linked channels if this has linke-enabled.
    guint i;
    for (i=0; i<initialized_ch; i++) {
      g_snprintf(tn, sizeof(tn), "voll%i", i); 
      w=lookup_widget(volumewindow,tn);
      if(gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(w))) {
	g_snprintf(tn, sizeof(tn), "voln%i", i); 
	w=lookup_widget(volumewindow,tn);
	port =  (gchar *) gtk_label_get_text(GTK_LABEL(w));
	//printf("SL-vol ch:%i n:%s - val %.3f\n",ch,port,val);
	set_port_volume(port,val);
      }
    }
  } else {
    g_snprintf(tn, sizeof(tn), "voln%i", ch); 
    w=lookup_widget(volumewindow,tn);
    port =  (gchar *) gtk_label_get_text(GTK_LABEL(w));
    //printf("SL-vol ch:%i n:%s - val %.3f\n",ch,port,val);
    set_port_volume(port,val);
  }
}


void update_volgui(guint ch) {
  guint i;
  gchar tn[32];
  GtkWidget *hbox; 

  if (ch == initialized_ch) return;

  hbox=lookup_widget(volumewindow,"hbox7");

  // TODO clear vol-gui
  for (i=0; i<initialized_ch; i++) {
    GtkWidget *w;
    MY_DESTROY_OBJECT ("vols%i"); 
    MY_DESTROY_OBJECT ("voln%i"); 
    MY_DESTROY_OBJECT ("voll%i"); 
    MY_DESTROY_OBJECT ("volb%i"); 
  }

  initialized_ch=0;
  for (i=0; i<ch; i++) {
    GtkWidget *vbox, *vscale, *lv_link, *lv_port;

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 0);
    vscale = gtk_vscale_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 100, 1, 0, 0)));
    gtk_scale_set_digits (GTK_SCALE (vscale), 0);
    gtk_range_set_inverted (GTK_RANGE (vscale), TRUE);
    gtk_box_pack_start (GTK_BOX (vbox), vscale, TRUE, TRUE, 0);
    gtk_scale_set_value_pos (GTK_SCALE (vscale), GTK_POS_BOTTOM);
    lv_link = gtk_check_button_new_with_mnemonic (_("link"));
    gtk_box_pack_start (GTK_BOX (vbox), lv_link, FALSE, FALSE, 0);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lv_link), TRUE);
    lv_port = gtk_label_new ("---");
    gtk_box_pack_start (GTK_BOX (vbox), lv_port, FALSE, FALSE, 0);
    gtk_widget_show (vbox);
    gtk_widget_show (vscale);
    gtk_widget_show (lv_link);
    gtk_widget_show (lv_port);
    g_signal_connect ((gpointer) vscale, "value_changed", G_CALLBACK (on_vol_value_changed), (void*)i); // XX
    MY_HOOKUP_OBJECT (volumewindow, vscale, "vols%i"); 
    MY_HOOKUP_OBJECT (volumewindow, lv_port, "voln%i"); 
    MY_HOOKUP_OBJECT (volumewindow, lv_link, "voll%i"); 
    MY_HOOKUP_OBJECT (volumewindow, vbox, "volb%i"); 

    initialized_ch++;
  }
}

void update_volume(guint ch, char* cname, guint value) {
  gchar tn[32];
  GtkWidget *w;
  g_snprintf(tn, sizeof(tn), "vols%i", ch);
  w=lookup_widget(volumewindow,tn);
  gtk_range_set_value (GTK_RANGE(w),value);
  g_snprintf(tn, sizeof(tn), "voln%i", ch);
  w = lookup_widget(volumewindow, tn); gtk_label_set_text(GTK_LABEL(w),cname);
}

/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * some gtk tree helpers
 */

/* look up position in playlist */
guint position_in_tree(GtkTreeModel *model, GtkTreePath *path ) {
  GtkTreeIter  iter;
  guint pos =0;

  if (!gtk_tree_model_get_iter_first(model, &iter)) { printf("empty list.\n"); return 0; } // actually an error!
  do {
    gboolean flag = FALSE;
    GtkTreePath *gtp=gtk_tree_model_get_path(model,&iter);
    if (gtk_tree_path_compare(gtp,path) == 0) { flag=TRUE;}
    gtk_tree_path_free(gtp);
    if (flag) break;
    pos++;
  } while (gtk_tree_model_iter_next(model, &iter));
  return pos;
}

void pls_mconnect(gboolean on) {
  GtkTreeView *tvp = GTK_TREE_VIEW(lookup_widget(mainwindow, "treeview1")); 
#if 0 //  this should work perfectly, but does not at all..
  static GdkRectangle rect = {0,0,0,0};
  if (on)  {
    gtk_tree_view_set_model (tvp, GTK_TREE_MODEL (lsp));
    printf("res-scroll: %i,%i,%i,%i\n",rect.x,rect.y,rect.width,rect.height);
    gtk_tree_view_scroll_to_point(tvp, rect.x, rect.y);
  } else  {
    gtk_tree_view_get_visible_rect(tvp,&rect);
    printf("sto-scroll: %i,%i,%i,%i\n",rect.x,rect.y,rect.width,rect.height);
    gtk_tree_view_set_model(tvp, NULL);
  }
#elif 1  // this might not work always
  static GtkTreePath *ps = NULL, *pe = NULL;
  if (on)  {
    gtk_tree_view_set_model (tvp, GTK_TREE_MODEL (lsp));
    if (ps) gtk_tree_view_scroll_to_cell(tvp, ps, NULL, true, 0.0, 0.1);
    //if (pe) gtk_tree_view_scroll_to_cell(tvp, pe, NULL, false, 0.0, 0.9);
  } else  {
    GtkTreeIter dummy;
    if (pe) gtk_tree_path_free(pe);
    if (ps) gtk_tree_path_free(ps);
    pe=ps=NULL;
    if (gtk_tree_model_get_iter_first(lsp, &dummy)) 
      gtk_tree_view_get_visible_range(tvp, &ps, &pe);
    gtk_tree_view_set_model(tvp, NULL);
  }
#else
  if (on)  {
    gtk_tree_view_set_model (tvp, GTK_TREE_MODEL (lsp));
    //expose_playlist(); 
  } else  {
    gtk_tree_view_set_model(tvp, NULL);
  }
#endif
}

void pls_scoll_to(guint pos) {
  guint i=0;
  GtkTreeIter iter;
  GtkTreeView *tvp = GTK_TREE_VIEW(lookup_widget(mainwindow, "treeview1")); 
  GtkTreeModel *model = GTK_TREE_MODEL(lsp);
  if (!gtk_tree_model_get_iter_first(model, &iter)) return; // actually an error!
  while (i<pos && gtk_tree_model_iter_next(model, &iter)) i++;
  //gtk_tree_model_iter_nth_child(model,&iter,NULL,pos);
  GtkTreePath *path = gtk_tree_model_get_path(model,&iter);
  gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(tvp), path, NULL, TRUE, 0.5, 0.5);
  gtk_tree_path_free(path);
}

void gx_plrm(guint pos_del) {
  gint pos=0;
  GtkTreeIter iter;
  GtkTreeModel *model = GTK_TREE_MODEL(lsp);
  if (!gtk_tree_model_get_iter_first(model, &iter)) return;
  do {
      if (pos == pos_del) {
	    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
	    break;
      }
      pos++;
  } while (gtk_tree_model_iter_next(model, &iter));
}

/* tree remove callback */
void del_sel_foreach(GtkTreeModel *tm, GtkTreePath *tp, GtkTreeIter *iter, gpointer data) {
  GtkTreePath *gtp=gtk_tree_model_get_path(tm,iter);
  printf("RM #%i\n",position_in_tree(tm,gtp));
  xmms_plrm(position_in_tree(tm,gtp));
  gtk_tree_path_free(gtp);
}


/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * seek slider
 */

int hsc_mode = 0;

void update_hsc (double val) {
  GtkWidget *scale;
  if (hsc_mode != 0 ) return;
  scale = lookup_widget(mainwindow,"hsc_seek");
  gtk_range_set_value (GTK_RANGE(scale),val);
}

/* vi:set ts=8 sts=2 sw=2: */
