/* 
   gx2ctl & mp3b - xmms2

   Copyright (C) 2007 Robin Gareus <robin@gareus.org>

   some handlers were inspired by gxmms2 - Johannes Heimansberg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/
//<-TERMINAL 124---------------------------------------------------------------------------------------------------------->/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "xmms2.h"
#include "trackinfo.h"

xmmsc_connection_t  *connection = NULL;

gchar *xmformat="${artist} - ${title}";
static trackinfo track;

static guint current_id = 0;
static guint last_dur = 0; 
static gint current_dur = 0; 
static guint curr_status = 0;

plsnfo my_playlist = { 0, 0, 0, 0 };

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
/** helpers */

gint res_has_key (xmmsc_result_t *res, const gchar *key) {
  return xmmsc_result_get_dict_entry_type (res, key) != XMMSC_RESULT_VALUE_TYPE_NONE;
}

gchar *format_url (gchar *item, GFileTest test) {
  gchar *url, rpath[PATH_MAX], *p;
  p = strchr (item, ':');
  if (!(p && p[1] == '/' && p[2] == '/')) {
    if (!realpath (item, rpath)) { return NULL; }
    if (!g_file_test (rpath, test)) { return NULL; }
    url = g_strdup_printf ("file://%s", rpath);
  } else {
    url = g_strdup_printf ("%s", item);
  }
  return url;
}

/* ----------------------- XMMS2 Volume ----------------------- */

typedef struct {
  guint chc;
  gint setvol;
} volst;

void set_port_volume(char *port, guint vol) {
  xmmsc_result_t *res;
  res = xmmsc_playback_volume_set (connection, port, vol);
  xmmsc_result_unref(res);
  //printf("set volume: %s = %d\n", port, vol);
}

void handle_volume_port (const void *key, xmmsc_result_value_type_t type, const void *value, void *userdata) {
  volst *vi = (volst*) userdata;
  if (vi->setvol <0 || vi->setvol >100) { // dump
    //printf("volume: %s = %u\n", (char*)key, XPOINTER_TO_UINT (value)); // FIXME
    update_volume(vi->chc,(char*)key,XPOINTER_TO_UINT (value));
  } else { //set
    set_port_volume((char*)key, vi->setvol);
  }
  ++(vi->chc);
}

void handle_count (const void *key, xmmsc_result_value_type_t type, const void *value, void *userdata) {
  ++(*((gint*)userdata));
}

static void  handle_volume_ports (xmmsc_result_t *res, void *userdata) {
  volst vi; 
  guint ch=0;
  static gboolean vlock = 0;
  if (!connection) {xmmsc_result_unref (res); return; } // workaround to disconnected xmms2 

  XMMS2CERR(res)
  vi.setvol= (guint)userdata;
  vi.chc= 0;

  if (vlock) return; vlock=true;
  xmmsc_result_dict_foreach (res, handle_count, &ch);
  vlock=false;
  update_volgui(ch);

  xmmsc_result_dict_foreach (res, handle_volume_port, &vi);
}

void cmd_volume (xmmsc_connection_t *conn, gint setvol) {
  xmmsc_result_t *res;
  res = xmmsc_playback_volume_get (conn);
  xmmsc_result_notifier_set (res, handle_volume_ports, (void*)setvol); // FIXME: we re-use the registered callback w/o unref
  xmmsc_result_unref (res);
}

/* ----------------------- internal fn & handlers ----------------------- */

static void handle_pls_list(xmmsc_result_t *res, void *userdata);
static void  handle_single_pls_entry (xmmsc_result_t *res, void *userdata);
static void do_mediainfo (xmmsc_result_t *res, void *userdata);
static void handle_current_id (xmmsc_result_t *res, void *userdata);

/** notifier: used to add playlists */
static void  handle_new_collection (xmmsc_result_t *res, void *userdata) {
  xmmsc_connection_t *conn = connection;
  gint pos = *((gint*)userdata);
  xmmsc_result_t *res2;
  xmmsc_coll_t *coll;
  const char *order[] = { "id", NULL };

  XMMS2CERR(res)
  if (!xmmsc_result_get_collection (res, &coll)) { XMMS2BROKEN(" - load playlist\n"); }
  if (pos>=0) res2 = xmmsc_playlist_insert_collection (conn, NULL, (guint) pos, coll, order); /// XXX x2-playlist name
  else res2 = xmmsc_playlist_add_collection (conn, NULL, coll, order); /// XXX x2-playlist name
  XMMS2CERR(res2)
  xmmsc_result_unref (res2);
  printf("Playlist with %d entries added\n", xmmsc_coll_idlist_get_size (coll));
  xmmsc_result_unref (res);
}

/** load and enque playlist (m3u/pls) */
static void add_playlist (xmmsc_connection_t *conn, gchar *plist, gint pos) {

  xmmsc_result_t *res;
  gchar *url;
  gint mypos =pos;

  url = format_url (plist, G_FILE_TEST_IS_REGULAR);

  res = xmmsc_coll_idlist_from_playlist_file (conn, url);
  xmmsc_result_notifier_set (res, handle_new_collection, &mypos);
  g_free (url);
  xmmsc_result_unref (res);
}

/** load and enqueue song */
void add_item_to_playlist (xmmsc_connection_t *conn, gchar *playlist, gchar *item, gint pos) {
  xmmsc_result_t *res;
  gchar *url;

  url = format_url (item, G_FILE_TEST_IS_REGULAR);
  if (!url) { fprintf (stderr, "Invalid url\n"); return; }

  if (pos>=0) res = xmmsc_playlist_insert_url (conn, playlist, (guint) pos, url);
  else res = xmmsc_playlist_add_url (conn, playlist, url);
  XMMS2CERR(res)
  xmmsc_result_unref (res);
  g_free (url);

  //printf("xmms2 added %s\n", item);
}

static void  handle_plpos (xmmsc_result_t *res, void *userdata) {
  guint ui;
  gboolean f = false;
  XMMS2CERR(res)
  xmmsc_result_get_uint (res, &ui);
  xmmsc_result_unref (res);
  if (!userdata) return;
  if ( *((guint *)userdata) != ui) { f= true; }
  *((guint *)userdata)=ui;
  if (f) {
    expose_playlist();
    //printf ("would scroll to now playing NOW!\n");
    if (0) pls_scoll_to(ui); // XXX make it optional
  }
}

void update_single_pls (guint id, gint pos_ins) {
  printf("update id:%i @pos:%i\n",id,pos_ins);
  xmmsc_result_t *res;
  res = xmmsc_medialib_get_info (connection, id);
  xmmsc_result_notifier_set (res, handle_single_pls_entry, (void *)pos_ins);
  xmmsc_result_unref(res);
}

/** updates which song is currently played - not it's actuall data */
void refresh_current (void) {
  xmmsc_result_t *res = xmmsc_playlist_current_pos (connection, NULL); // XXX x2-playllist name
  xmmsc_result_notifier_set (res, handle_plpos, &(my_playlist.playlist_pos));
  xmmsc_result_unref (res);
}

void refresh_trackinfo (void) {
  if (!current_id) return;
  //printf("refreshing info for id %i\n",current_id);
  xmmsc_result_t *res = xmmsc_medialib_get_info (connection, current_id);
  xmmsc_result_notifier_set (res, do_mediainfo, NULL);
  xmmsc_result_unref (res);
}

void refresh_current_id (void) {
  //printf("refreshing current id.\n");
  xmmsc_result_t *res = xmmsc_playback_current_id (connection);
  xmmsc_result_notifier_set (res, handle_current_id, NULL); //< will call a do_mediainfo (beware! of loops)
  xmmsc_result_unref (res);
}

void refresh_playlist(void) {
  refresh_current();
  xmmsc_result_t *res;
  res = xmmsc_playlist_list_entries (connection, NULL); /// XXX x2-playlist name
  xmmsc_result_notifier_set (res, handle_pls_list, &my_playlist);
  xmmsc_result_unref(res);
}

/* -----------------------  handlers for callbacks ----------------------- */

#include <gnome.h>
extern GtkListStore *lsp; ///< playlist list store
static gboolean pls_update_active;

static void  handle_coverart (xmmsc_result_t *res, void *userdata) {
  gchar *blob;
  size_t len=0;
  xmmsc_result_get_bin(res, (void*) &blob, &len);
  //printf("cover %i bytes\n",len);
  update_coverart(NULL, blob,len);
  xmmsc_result_unref (res);
}

//* same as below .../*
static void  handle_single_pls_entry (xmmsc_result_t *res, void *userdata) {
  trackinfo ti;
  gint pos_ins = (gint)userdata;
  GtkTreeIter iter;
  pls_mconnect(false);
  if (pos_ins<0) 
    gtk_list_store_append(lsp, &iter);
  else
    gtk_list_store_insert(lsp, &iter, pos_ins);
  trackinfo_update(res, &ti);
  ls_update(&iter,&ti);
  pls_mconnect(true);
  // TODO adjust total playtime here ?
  ++(my_playlist.playlist_length);
  xmmsc_result_unref(res);
}

//*  FIXME rename fn- does no longer (only) sum up pls duration..
static void  handle_pls_time (xmmsc_result_t *res, void *userdata) {
  trackinfo ti;
  gint playtime;
  GtkTreeIter iter;
  plsnfo *p = (plsnfo*) userdata;

  trackinfo_update(res, &ti);

  // TOTAL PLAYTIME

  if (trackinfo_has_playtime(&ti))
    p->total_playtime += trackinfo_get_playtime(&ti);

  // LISTSTORE
  if (lsp != NULL) {
    gtk_list_store_append(lsp, &iter);
    ls_update(&iter,&ti);
  }

  if (++(p->process_tottime) == p->playlist_length) {  // LAST
    status_printf("total playtime: %.1f min", p->total_playtime/60000.0);
    pls_mconnect(true);
    pls_update_active=false;
  }
  xmmsc_result_unref(res);
}

static void handle_pls_list(xmmsc_result_t *res, void *userdata) {
  plsnfo *p = (plsnfo*) userdata;
  if (pls_update_active) return; 
  pls_update_active=true;
  p->playlist_length=p->total_playtime=p->process_tottime=0;

  //printf("DEBUG PLS-LST\n");
  pls_mconnect(false);
  gtk_list_store_clear(lsp);
  for (; xmmsc_result_list_valid(res); xmmsc_result_list_next(res)) {
    xmmsc_result_t *info_res;
    guint ui;
    if (!xmmsc_result_get_uint (res, &ui)) { XMMS2BROKEN("") }
    // ---
    //p->pls2d = realloc(p->pls2d, (p->playlist_length+1) * sizeof(guint));
    //p->pls2d[p->playlist_length] = ui;
    // ---
    info_res = xmmsc_medialib_get_info (connection, ui);
    xmmsc_result_notifier_set (info_res, handle_pls_time, p);
    xmmsc_result_unref(info_res);
    p->playlist_length++;
  }
  if (!p->playlist_length) pls_update_active=false;
  xmmsc_result_unref(res);
}

static void do_mediainfo (xmmsc_result_t *res, void *userdata) {
  XMMS2CERR(res)
  trackinfo_update(res, &track);
  //printf("DEBUG DO-MI\n");
  if (trackinfo_has_playtime(&track)) {
    current_dur=trackinfo_get_playtime(&track);
  } else {
    current_dur = 0; 
  }
  update_status(curr_status, current_dur); 

  update_info(&track);
  refresh_current();

  if (trackinfo_has_picture(&track)) {
    xmmsc_result_t *res2;
    //printf("DEBUG: pic-hash: %s\n",trackinfo_get_picture_hash(&track));
    res2=xmmsc_bindata_retrieve(connection,trackinfo_get_picture_hash(&track));
    xmmsc_result_notifier_set (res2, handle_coverart, NULL);
    XMMS2CERR(res2)
    xmmsc_result_unref (res2);
  } else {
    wipe_coverart();
  }

  xmmsc_result_unref (res);
}

/* ----------------------- registered callback handlers ----------------------- */

static void handle_playtime (xmmsc_result_t *res, void *userdata) {
  xmmsc_result_t *newres;
  guint dur;
  if (!connection) {xmmsc_result_unref (res); return; } // workaround to disconnected xmms2 

  XMMS2CERR(res)
  if (!xmmsc_result_get_uint (res, &dur)) { XMMS2BROKEN("") }

  //if [stream] update track-info (current_id) every 5s 
  if (connection &&  trackinfo_is_stream(&track) && (((dur / 5000) % 10) != ((last_dur / 5000) % 10))) {
    //printf("[stream] refresh %i\n",dur);
    refresh_current_id();
  }

  newres = xmmsc_result_restart (res);
  if (((dur / 100) % 20) != ((last_dur / 100) % 20)) {
    last_dur = dur;
    //printf("xmms: duration - %i/%i\n",last_dur,current_dur);
#if 1
    double v= 1000.0*dur; v/=current_dur; 
    update_hsc(v);
    update_infotime(dur, current_dur);
#endif
  }
  xmmsc_result_unref (res); // this should stop cont. updates, but it works just fine.
  xmmsc_result_unref (newres);
}

static void handle_current_id (xmmsc_result_t *res, void *userdata) {
  if (!connection) {xmmsc_result_unref (res); return; } // workaround to disconnected xmms2 
  //printf("DEBUG - handle current_id\n");
  XMMS2CERR(res)
  if (!xmmsc_result_get_uint (res, &current_id)) { XMMS2BROKEN("") }

  if (current_id) {
    res = xmmsc_medialib_get_info (connection, current_id);
    xmmsc_result_notifier_set (res, do_mediainfo, NULL);
    xmmsc_result_unref (res);
  }
}

static void handle_status_change (xmmsc_result_t *res, void *userdata) {
  guint new_status;
  if (!connection) {xmmsc_result_unref (res); return; } // workaround to disconnected xmms2 
  XMMS2CERR(res)
  if (!xmmsc_result_get_uint (res, &new_status)) { XMMS2BROKEN("") }
  //fprintf(stdout, "xmms2: status %i -> %i\n",curr_status, new_status);
  //status_printf("xmms2: status %i -> %i",curr_status, new_status);
  curr_status = new_status;
  update_status(new_status, current_dur); 
}


static void handle_xmms_disconnect (void *userdata) {
  xmmsc_unref(connection);
  connection=NULL;
  status_printf("xmms2: lost connection");
  update_gui(); 
}


static void handle_playlist_change (xmmsc_result_t *res, void *userdata) {
  GtkTreeIter   iter;
  trackinfo ti;
  gint signal, id;
  gint pl_length, pos_del = -1, pos_ins = -1, pos = -1;
  gchar *title;
  if (!connection) {xmmsc_result_unref (res); return; } // workaround to disconnected xmms2 
  XMMS2CERR(res)

  xmmsc_result_get_dict_entry_int(res, "type", &signal);

  refresh_current();

  if (!lsp) { return; }

  //printf("DEBUG Playlist changed: %i\n",signal);

  switch (signal) {
    case XMMS_PLAYLIST_CHANGED_ADD:
    #if 1
      //printf("DEBUG ADD\n");
      xmmsc_result_get_dict_entry_uint(res, "id", &id);
      update_single_pls(id,-1);
      break;
    #endif
    case XMMS_PLAYLIST_CHANGED_INSERT:
    #if 1
      //printf("DEBUG INSERT\n");
      xmmsc_result_get_dict_entry_uint(res, "id", &id);
      xmmsc_result_get_dict_entry_int(res, "position", &pos_ins);
      update_single_pls(id,pos_ins);
      break;
    #endif
    case XMMS_PLAYLIST_CHANGED_SHUFFLE:
    case XMMS_PLAYLIST_CHANGED_SORT:
      refresh_playlist();
      break;
    case XMMS_PLAYLIST_CHANGED_REMOVE:
      xmmsc_result_get_dict_entry_int(res, "position", &pos_del);
      //printf("DEBUG REMOVE %i\n",pos_del);
      //TODO re-calc total play time.
      gx_plrm(pos_del);
      break;
    case XMMS_PLAYLIST_CHANGED_CLEAR:
      gtk_list_store_clear(lsp);
      break;
    case XMMS_PLAYLIST_CHANGED_MOVE:
      //printf("DEBUG CHANGED MOVE\n");
      refresh_playlist();
      break;
  }

}

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
/* ----------------------- public functions - gx2ctl API ----------------------- */

int init_xmms_session (void) {
  gchar *path = NULL; // TODO: get from prefs.
  gint conn_status;
  if(connection) {
    fprintf(stderr,"xmms2 already connected.\n");
    return 1;
  }

  if (!(connection = xmmsc_init("gx2ctl"))) {
    fprintf (stderr, "Could not init xmmsc_connection!\n");
  }

  if (!(path = getenv("XMMS_PATH"))) {
    if (path == NULL) asprintf(&path, "unix:///tmp/xmms-ipc-%s", g_get_user_name());
  }

  conn_status = xmmsc_connect(connection, path);

  if (!conn_status && 1 ) {
    if (!path || g_ascii_strncasecmp(path, "unix://", 7) == 0) {
      fprintf (stdout, "starting up xmms2 daemon.\n");
      if (!system("xmms2-launcher")) {
	conn_status = xmmsc_connect(connection, path);
      }
    }
  }

  if (!conn_status) {
    fprintf(stderr, "Could not connect to xmms2d: %s\n", xmmsc_get_last_error (connection));
    xmmsc_unref (connection);
    connection=NULL;
    return (0);
  } 

  XMMS_CALLBACK_SET (connection, xmmsc_broadcast_playback_current_id, handle_current_id, NULL);
  XMMS_CALLBACK_SET (connection, xmmsc_broadcast_playback_status, handle_status_change, NULL);
  XMMS_CALLBACK_SET (connection, xmmsc_playback_current_id, handle_current_id, NULL);
  XMMS_CALLBACK_SET (connection, xmmsc_playback_status, handle_status_change, NULL);
  XMMS_CALLBACK_SET (connection, xmmsc_broadcast_playlist_changed, handle_playlist_change, NULL);
  XMMS_CALLBACK_SET (connection, xmmsc_signal_playback_playtime, handle_playtime, NULL);
  XMMS_CALLBACK_SET (connection, xmmsc_broadcast_playback_volume_changed, handle_volume_ports, (void*) -1);

  xmmsc_disconnect_callback_set (connection, handle_xmms_disconnect, NULL); 
  xmmsc_mainloop_gmain_init(connection); 

  refresh_playlist();
  refresh_current();
  //cmd_volume(connection,100);
  cmd_volume(connection,-1);
  pls_update_active=false;
  status_printf("xmms2: connected");
  return (1);
}

#define XHD(meth) res=meth(connection); xmmsc_result_disconnect(res); 
#define XHU(meth) res=meth(connection); xmmsc_result_unref(res); 

void close_xmms_session (void) {
  if(!connection) return;
#if 0
  xmmsc_result_t *res;
  XHD(xmmsc_broadcast_playback_volume_changed)
  XHD(xmmsc_broadcast_playlist_changed)
  XHU(xmmsc_broadcast_playback_current_id)
  XHU(xmmsc_broadcast_playback_status)
  XHU(xmmsc_signal_playback_playtime)
#endif
#if 1
  xmmsc_result_t *res;
  xmmsc_connection_t  *tmp = connection;
  xmmsc_disconnect_callback_set (connection, NULL, NULL); 
  connection=NULL;
  // workaround: provoke callbacks w/ NULL connection -> unref/disconnect! 
  res = xmmsc_playlist_move_entry (tmp, NULL, 0, 0); ///does not work on empty playlist
  XMMS2CERR(res) xmmsc_unref (tmp);
  //similarly query&set volume 
  //and if not playing: time/duration - better toggle status 2x. ;(
#else
  xmmsc_unref (connection);
  connection=NULL;
#endif
}

void xmms2_cmd (int cmd) {
  if (!connection) return;
  xmmsc_result_t *res;

  switch (cmd) {
    case XCMD_PLAY:
      res = xmmsc_playback_start (connection); break;
    case XCMD_PAUSE:
      res = xmmsc_playback_pause (connection); break;
    case XCMD_STOP:
      res = xmmsc_playback_stop (connection); break;
    case XCMD_NEXT:
      res = xmmsc_playlist_set_next_rel (connection, 1);
      XMMS2CERR(res)
      xmmsc_result_unref (res);
      res = xmmsc_playback_tickle (connection);
      break;
    case XCMD_CURR:
      res = xmmsc_playlist_set_next_rel (connection, 0);
      XMMS2CERR(res)
      xmmsc_result_unref (res);
      res = xmmsc_playback_tickle (connection);
      break;
    case XCMD_FIRST:
      res = xmmsc_playlist_set_next (connection, 0);
      XMMS2CERR(res)
      xmmsc_result_unref (res);
      res = xmmsc_playback_tickle (connection);
      break;
    case XCMD_LAST:
      res = xmmsc_playlist_set_next (connection, my_playlist.playlist_length-1); 
      XMMS2CERR(res)
      xmmsc_result_unref (res);
      res = xmmsc_playback_tickle (connection);
      break;
    case XCMD_PREV:
      res = xmmsc_playlist_set_next_rel (connection, -1);
      XMMS2CERR(res)
      xmmsc_result_unref (res);
      res = xmmsc_playback_tickle (connection);
      break;
    case XCMD_FRWD:
      res = xmmsc_playback_seek_ms(connection, last_dur-5000); // skip 5s
      break;
    case XCMD_FFWD:
      res = xmmsc_playback_seek_ms(connection, last_dur+5000); // skip 5s
      break;
    case XCMD_CLEAR:
      res = xmmsc_playlist_clear (connection, NULL); // XXX x2-playlist name
      break;
    default:
      fprintf(stderr, "unknown xmms2 command\n");
      return;
  }
  XMMS2CERR(res)
  xmmsc_result_unref (res);
}

void add_to_playlist (char *fn, gint pos) {
  char *tmp;
  if (!connection) return;
  if ((tmp=strchr(fn,'\n'))) *tmp=0;
  int ln=strlen(fn);
  if (ln>4 && (// match /^.*\.{pls|m3u}$/ 
    !(strncmp(".m3u",&(fn[ln-4]),4)) || !(strncmp(".pls",&(fn[ln-4]),4))
	      )) 
    add_playlist(connection, fn, pos);
  else 
    add_item_to_playlist(connection, NULL, fn, pos); // XXX set x2-playlist name
  //ml_lookup(fn); // XXX medialib 
}

void xmms_reposition (float n) { 
  xmmsc_result_t *res;
  guint ms = n/1000.0 * current_dur;
  res = xmmsc_playback_seek_ms(connection, ms); 
  XMMS2CERR(res)
  xmmsc_result_unref (res);
}

void xmms_plrm (guint pos) { 
  xmmsc_result_t *res;
  res = xmmsc_playlist_remove_entry (connection, NULL, pos); /// XXX x2-playlist name
  XMMS2CERR(res)
  xmmsc_result_unref (res);
}

void xmms_plmv (guint from, guint to) { 
  xmmsc_result_t *res;
  res = xmmsc_playlist_move_entry (connection, NULL, from, to); /// XXX x2-playlist name
  XMMS2CERR(res)
  xmmsc_result_unref (res);
}

void xmms_pljump (guint n, gboolean start_play) { 
  xmmsc_result_t *res;

  if (n >= my_playlist.playlist_length) { fprintf(stderr, "selection error\n"); return; }
  res = xmmsc_playlist_set_next (connection, n);
  XMMS2CERR(res)
  xmmsc_result_unref (res);
  if (curr_status ==0 && start_play)  xmms2_cmd(XCMD_PLAY);
  else {
    if (curr_status ==2 && start_play) xmms2_cmd(XCMD_PLAY);
    res = xmmsc_playback_tickle (connection);
    XMMS2CERR(res)
    xmmsc_result_unref (res);
  } 
}

int xmms_is_connected (void) { return(connection?1:0); }

/* vi:set ts=8 sts=2 sw=2: */
